/* 
 * Thomas Pantzer: t_polygon.cc, v0.1b  8.11.98 0:58 
 * 
 * 
 * Copyright 1998 by Thomas Pantzer
 * 
 * Permission to use, copy, modify, and distribute this software for noncommercial 
 * use and without fee is hereby granted, provided that the above copyright notice 
 * appear in all copies and that both that copyright notice and this permission 
 * notice appear in supporting documentation, and that the name of the Author not be 
 * used in advertising or publicity pertaining to distribution of the software 
 * without specific, written prior permission. The Author makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is" 
 * without expressed or implied warranty. 
 * 
 * 
 * pantec@informatik.uni-leipzig.de   (Thomas Pantzer) 
 * 
 */



#define _T_POLYGON_G
#include "t_polygon.h"
#include "t_line.h"

//module definition and procedures go here

void t_polygon::Init(t_colormap *c, t_vector2d *a, t_drawobj *n, int lw, int s)
{
  t_drawgrp::Init(c, a, n, lw);
  kind = _polygon;
  Shape = s;
  filled = false;
}

void t_polygon::Draw(color_type ct)
{
#ifdef DEBUG
   fprintf(stderr, "t_polygon::Draw()\n");    
#endif   
   t_drawgrp::Draw(ct);

   if (filled == true)
      Fill();
}

void t_polygon::AddLine(t_vector2d *P)
{
   t_line *L = new t_line;
   t_vector2d *A;
   
   if (first == NULL)
      A = &anchor;
    else
      {
       t_line *ll = (t_line *)first;
       A = &(ll->endpoint);
      } 
      
   L->Init(cmap,A,first,P,width);
   first = L;
   NewAnchor();
   count++;
}

void t_polygon::ChangeLine(t_vector2d *P)
{
   if (first == NULL)
      return;
      
   t_line *LL = (t_line *)first;
   LL->endpoint.Init(P);
}


void t_polygon::Fill(void)
{
   t_drawobj *p = first;
   XPoint *point;
   
   if (count == 0)
      return;
   
   point = new XPoint[count];

   int i = 0;
   while (p != NULL)
     {
      point[i].x = p->anchor.GetX();
      point[i].y = p->anchor.GetY();
      p = p->next;
      i++;
     }
     
   FillPoly(point,i,Shape);  

   delete point;
}


		       
void t_polygon::Done(void)
{
      t_line **LP,*p;
      
      LP = new t_line * [count * 2];
      p = (t_line *)first;
      int i = 0;
      while (p != NULL)
        {
         LP[i] = p;
         i++;
         p = (t_line *)p->next;
        }
        
      for (int j=0; j<i; j++)
        delete LP[j];
        
      delete [] LP;    

      first = NULL;
}

void t_polygon::SaveMeToo(int fd)
{
   t_drawgrp::SaveMeToo(fd);

   char st[256];
   sprintf(st,"\tfl:%d\tos:%d",filled,Shape);
   write(fd,st,strlen(st));
}



