/*
 *  linux/drivers/char/scanner.c
 *
 *  Copyright (c) 1994 Thomas Faehnle  (Thomas.Faehnle@student.uni-ulm.de)
 *
 *  This is the driver for the M105 400 dpi halftone handheld scanner driven 
 *  thru a little interface using the GI1904 chip. IRQ and DMA channel are
 *  adjusted thru software, they are defined in scanner.h.
 *  Note: This scanner is sold under various names, usually ending in "105",
 *	  the original manufacturer is Mustek.
 */

/* 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  m105.c,v 1.8 1994/04/13 22:28:42 blurf Exp
 */
 
#include "m105.h"


#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>	 /* for f_flags bits */
#include <linux/module.h>
#include <linux/ioport.h> 
#define REALLY_SLOW_IO
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/mm.h>
/* #define __NO_VERSION__ */

#include <linux/version.h>

/* I made this with kernel versions 2.2.5, 2.0.36  and one somewhat older (I don't remember) */

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
	#include <asm/uaccess.h>
	#include <linux/poll.h>
	#include <linux/interrupt.h>
	#include <asm/spinlock.h>

	#define memcpy_fromfs(X,Y,Z)  copy_from_user(X,Y,Z)
	#define memcpy_tofs(X,Y,Z)    copy_to_user(X,Y,Z)

   	#include <linux/proc_fs.h>
   	#include <linux/miscdevice.h>
extern struct proc_dir_entry proc_m105;

/*
extern int  m105_proc_init(void);
extern void m105_proc_destroy(void);
*/
#else
   	#include <asm/segment.h>
#endif


#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
   	MODULE_PARM(iobase,"1i");
        static int iobase = 0x2ab

   	MODULE_PARM(irq,"1i");
        static int irq = 0;
        
   	MODULE_PARM(dma,"1i");
        static int dma = 0;  

   	MODULE_PARM(msize,"1i");
        static int msize = (64+32+16);  

   	MODULE_PARM(auto_irq,"1i");
        static int auto_irq = 1;  

        #define GI1904_IOBASE iobase
        #define GI1904_DMA    dma
        #define GI1904_IRQ    irq	

   	static int irq_list[3] = {10,5,3}; 
        static int dma_list[2] = {3,1}; 
	
	rwlock_t rw_lock = RW_LOCK_UNLOCKED;
	
#endif
        


#define _modeindex_	(SScanner[SCD_NUM].modeindex)
#define _MODEINFO_	(SScanner[SCD_NUM].mode[_modeindex_])
#define OPERATE 	(SScanner[SCD_NUM].op_cmd)



#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))

/*
#define RD_BLKEND	(rd.blkptr->blkoff + rd.blkptr->blksiz)
#define WR_BLKEND	(wr.blkptr->blkoff + wr.blkptr->blksiz)
#define RD_BUFSPC	(wr.index >= rd.index ? ((wr.index == rd.index) && !ds.buffer_empty ? \
	bufsiz : 0) + wr.index - rd.index : bufsiz - rd.index + wr.index)
#define WR_BUFSPC	(wr.index < rd.index ? rd.index - wr.index : \
	((wr.index == rd.index) && !ds.buffer_empty ? 0 : bufsiz) - wr.index + rd.index)

#define RD_ADDR		(&(rd.blkptr->data[rd.index - rd.blkptr->blkoff]))
#define WR_ADDR		(&(wr.blkptr->data[wr.index - wr.blkptr->blkoff]))

#define SELECT_TH	(_MODEINFO_.bpl*_MODEINFO_.depth*_MODEINFO_.sc_mode * select_threshold)
*/

#ifdef DEBUG_M105
#define PRINTK(x...)	{ if (debug) printk(x); }
#else
#define PRINTK(x...)
#endif


#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0))
   static wait_queue_head_t wq;
#else
   static struct wait_queue *wq = NULL;
#endif
   
/* static int modeindex = 0; */

static int SCD_NUM = 0;
static struct device_status_struct ds = { 0, 0, 0, 0 };

/*
static struct buffer_block_struct *memroot = NULL;
static int buflines = 0, req_lines = 0, bufsiz = 0, bytes_missing = 0;
static struct buffer_pointer rd = { NULL, 0 }, wr = { NULL, 0 };
static int select_threshold = 1;
*/

static char M105_ID[] = "m105";

static long real_bufsize = 0;

#ifdef DEBUG_M105
static int debug = 1; 
#endif

static long int _BPL_ = 0;
static long int old_dma_count = 0;

char kernel_version[]=UTS_RELEASE;	/* needed for module installer */


int intr_count = 0;

#include "queue_buffer.c"


static __inline__ int dma_terminal_count(unsigned int channel)
{
  if (channel <= 3)
    return inb(DMA1_STAT_REG) & (1 << channel);
  else
    return inb(DMA2_STAT_REG) & (1 << (channel & 3));
}

/* This function is evil, evil, evil.  This should be a
   _kernel_ rescheduling sleep! */
static void __cold_sleep(int tics)	/* stolen from linux/drivers/net/plip.c */
{
    int start = jiffies;
    while(jiffies < start + tics)
	; /* do nothing */
    return;
}



int __getmode_low(void)
{
	unsigned char b1,b2;
	int i;
	
	b1 = 0;
	b2 = 0;

        for (i=0; i<20000; i++)
	  {
	   b2 = b1;
 	   b1 = inb(GI1904_MISC2);

	   if ((b1 & 0x3f) == 0x27)	
	      {
	       unsigned int code = ((((b1 & 0xc0) >> 2) | (b2 & 0xc8)) >> 1) | 0x80;
	       return code;
	      }	
	  }		  
	
	return -1;	  
}


int getmode(void)
{
	int A,B,C,i;

	outb(0xe3,GI1904_CONTROL); 

	A = 0;
	B = 0;
	C = 0;
        for (i=0; i<20000; i++)
	  {
	   C = B;
	   B = A;
	   A = __getmode_low();

/*	   usleep(10000);  */
	   __cold_sleep(2);		   
	   
	   if (A < 0)
	      {
	       outb(0x0,GI1904_CONTROL); 
	       return A; 
	      } 
	   
	   if ((A == B) && (B == C))
	      {
	       outb(0x0,GI1904_CONTROL); 
	       return A;
	      } 
	  }

	return -1;	  
}

static int m105_get_mode(void)
{
	int mode,n,i;
/*
	outb(0x91,GI1904_CONTROL); 
	if ( !((mode = inb_p(GI1904_CONTROL)) & 0x80) ) 
            mode &= GI1904_MODE_MASK;
*/

	mode = getmode();
	
	if (mode < 0)
	   return -ENODEV;
	   
	n = 0;
        while (SScanner[n].id != 0)
          {
           for (i=0; i<SScanner[n].num_modes; i++)
             {
	      if ((mode & 0xc7) == 0xc4)	
	         mode &= 0xef;
		 
              if (SScanner[n].mode[i].detect == mode)
                 {
                  if (SScanner[n].mode[i].options == SCC_SOFTDETECT)
                     {
                      SScanner[n].modeindex = i;
                      SCD_NUM = n;
                      return 0;
                     } 
                 } /* if */
             } /* for */
           n++;  
          } /* while */
	
	return -EIO;
}


static int m105_get_mode__(void)
{
  int n,k, i, mode; 

  for(k=0; k<10000; k++) 
    { 
     /* scanner on to read back scanner identity */ 
     outb(0x91,GI1904_CONTROL); 

     __cold_sleep(10);
     if ( !((mode = inb_p(GI1904_CONTROL)) & 0x80) ) 
        {
         mode &= GI1904_MODE_MASK;
         
         PRINTK("m105: regs dump: r1:0x%03x r2:0x%03x r3:0x%03x\n",
                  inb_p(GI1904_MISC0),inb_p(GI1904_MISC1),inb_p(GI1904_MISC2));
         
         n = 0;
         while (SScanner[n].id != 0)
           {
            for (i=0; i<SScanner[n].num_modes; i++)
              {
               if (SScanner[n].mode[i].detect == mode)
                  {
/*
                   if (SScanner[n].mode[i].options == SCC_SOFTDETECT)
                      {
 */
                       SScanner[n].modeindex = i;
                       SCD_NUM = n;
                       return 0;
/*
                      } 
 */                      
                  } /* if */
              } /* for */
            n++;  
           } /* while */

/*
         for(n = 0; n<(sizeof(m105_modebits) / sizeof(m105_modebits[0])); n++)
   	   if (m105_modebits[n] == mode) 
              {
	       modeindex = n;
               PRINTK("m105: modeindex %d\n",modeindex);
	       return 0;
	      }  
 */
              
        } /* if */     
        
    } /* for */
    
  return -EIO;
}

static __inline__ void setup_xfer(char *addr, long int BPL)
{
	unsigned long dma_lock_flags;

  	dma_lock_flags = claim_dma_lock();

/*	disable_dma(GI1904_DMA); */

	set_dma_mode(GI1904_DMA, DMA_MODE_READ);
	clear_dma_ff(GI1904_DMA);
	set_dma_addr(GI1904_DMA, (int) addr);
	set_dma_count(GI1904_DMA, BPL);
	old_dma_count = BPL;

	enable_dma(GI1904_DMA);

	release_dma_lock(dma_lock_flags);
}


static __inline__ void scanner_standby(void)
{
	unsigned long dma_lock_flags;

	ds.irq_enabled = 0;  

	outb(STANDBY|OPERATE, GI1904_CONTROL);	/* gives me a safer feeling if the */

  	dma_lock_flags = claim_dma_lock();
	disable_dma(GI1904_DMA);		/*  DMA channel is also disabled on the 8237 */
	release_dma_lock(dma_lock_flags);
}

static __inline__ void scanner_on(void)
{
	setup_xfer(&Queue.wptr->data[Queue.wptr->filled],_BPL_);
	outb(OPERATE, GI1904_CONTROL);

	ds.irq_enabled = 1;
} 


static __inline__ void scanner_start(void)
{
	_BPL_ = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth);
        Queue.wptr->filled = 0;
        Queue.wptr->gotten = 0;
        Queue.rptr = Queue.wptr;
        
	scanner_on();
}


static __inline__ int read_queue(char **user_buf, int requested)
{
/*   _BPL_ = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth); */
  int avail,ofs,extra,got;
  char *dest;  

      dest = *user_buf;
      got = 0;
      extra = 0;

      avail = bytes_in_queue();

      while (avail > 0)
	 {
	  ofs = Queue.rptr->gotten;

          cli();
	  memcpy_tofs(&dest[got], &Queue.rptr->data[ofs], avail);
	  Queue.rptr->gotten += avail;
	  got += avail;
          sti();
             
    /*       PRINTK("m105: read_queue(): read stack is %ld, extra %d\n",Queue.rptr->gotten,extra);

	  if (Queue.rptr->gotten + (_BPL_*3) > Queue.rptr->size)
             {
             }
     */         
	  avail = bytes_in_queue();   
	  extra++;
	 }  
  
  return got;
}



#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
static int m105_read(struct inode *inode, struct file *file, char *buf, int requested)
{
#else
static ssize_t m105_read(struct file *file, char *buf, size_t requested, loff_t *unknown)
{
    struct task_struct *Curr; 
#endif
    int r, ofs,rd,req;

  req = requested;

  if (!ds.irq_enabled)
     {
      scanner_start();
      PRINTK("m105: read: scanner on bpl = %ld\n",_BPL_);
     } 
 
  if (file->f_flags & O_NONBLOCK) {
     PRINTK("m105_read: reading in O_NONBLOCK mode\n");
  
     r = read_queue(&buf, requested);
     return r;
  }

  ofs = 0;
  rd = 0;
  while (rd < requested)
    {
     r = read_queue(&buf,req); 
     if (r > 0)
        {
         rd += r;
         ofs = rd;
         req -= r;
        }
        
     if (rd == requested)
        break;
        
    if (r <= 0)
       {
        PRINTK("m105: read(): %d bytes are still missing," 
                          "I'll sleep a bit. ds.irq_enabled == %d\n",req,ds.irq_enabled);
                           
        ds.process_sleeps = 1;
        interruptible_sleep_on(&wq);
        current->state = TASK_RUNNING;
       } 

/* #define SIGNAL_OCCURED	(current->signal & (~current->blocked)) */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
    if (current->signal & ~current->blocked)
#else
    Curr = get_current();    
    if (Curr->signal.sig[0] & ~Curr->blocked.sig[0]) 
/*    | (Curr->signal.sig[1] & ~Curr->blocked.sig[1]))  */
#endif
        return rd;
   }
        

  return requested;
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
static int m105_select(struct inode *inode, struct file *file, int type, select_table *st)
#else
static unsigned int m105_select(struct file *file, struct poll_table_struct *st) 
#endif
{
   	int avail;
  
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
     	if (type != SEL_IN)
           return 0;
#endif
/*  _BPL_ = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth); */

	avail = bytes_in_queue();

	if (avail > 0) {
	   PRINTK("m105: select: device is ready %d\n",avail);
	   return 1;			/* device is ready */
	}

	if (!ds.irq_enabled) {
	   PRINTK("m105: select: IRQ not enabled -- will turn on scanner\n");
	   scanner_start();         
	}                                               
                                              
	if (st) {
/*    	   PRINTK("m105_select: waiting for device to get ready\n");
*    bytes_missing = SELECT_TH; */
	   ds.process_sleeps = 1;

	   if (!ds.irq_enabled) {
	      PRINTK("m105: select: IRQ not enabled -- will turn on scanner, "
	        "waiting for device to get ready\n");
	      scanner_start();
	   }
	 else
	   {
	    PRINTK("m105: select: device NOT ready\n");
	   }    
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
	select_wait(&wq, st);
#else
	poll_wait(file, &wq, st); 
#endif    
	return 0;
}



#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,0,33))
static void m105_interrupt(int unused_pointer_to_pt_regs_struct)
#else
static void m105_interrupt(int irq_dummy,void *dev_id,struct pt_regs *unused_pointer_to_pt_regs_struct)
#endif
{
	unsigned long flags,misc;
	unsigned char b[8];
	int i;
	
	save_flags(flags);
	cli();	

	if (!ds.irq_enabled) {
	   restore_flags(flags);
	   PRINTK("scanner.c: m105_interrupt: unexpected interrupt %d\n",irq_dummy);
	   return;
	}

	if (!dma_terminal_count(GI1904_DMA)) {
	   restore_flags(flags);
	   printk("scanner.c: m105_interrupt: DMA transfer not finished yet\n");
	   return;
	}

 	intr_count++;	
	/* check interface for scan ready */

/*
	misc = 0;
	for (i=0; i<4; i++)
	  {
	   b[i] = inb(GI1904_MISC2);
	   misc <<= 8; 
	   misc |= b[i];
	  } 

	printk("m105: [% 8d] reg.m2 0x%08lx\n",intr_count,misc);
*/
/*
	for (i=0; i<200000; i++)
	  {
	   b[0] = inb_p(GI1904_MISC2);
	   b[1] = inb_p(GI1904_MISC2);
	   b[2] = inb_p(GI1904_MISC2);
	   
	   if ((b[0] == 0x77) && (b[1] == 0x73) && (b[2] == 0x73))
	      break;	
	  }	
*/
	  
// 	b[0] = inb_p(GI1904_MISC2); 
//	if ((inb_p(GI1904_MISC2) & 0x08) > 0)
//	   printk("m105: [% 8d] broken line\n",intr_count); 	

	scanner_standby(); 

/*
    PRINTK("m105: regs dump: r0: 0x%03x r1:0x%03x r2:0x%03x r3:0x%03x\n",
                  inb_p(GI1904_CONTROL),inb_p(GI1904_MISC0),
                  inb_p(GI1904_MISC1),inb_p(GI1904_MISC2));


	printk("m105_interrupt:");

	 r0: 0x%02x ",inb_p(GI1904_CONTROL));
	
	for (i=0; i<8; i++)
	   printk("%02x,",inb_p(GI1904_MISC0));
	  
	printk("\t");

	for (i=0; i<32; i++)
	   printk("%01x",(inb_p(GI1904_MISC2) & 0xf) >> 3);
	  
	printk("\n");
	         
    ds.irq_enabled = 0; */
/*    cli(); */

    	Queue.wptr->filled += old_dma_count;
    
	if ((Queue.wptr->size - Queue.wptr->filled) < (_BPL_*4) )
	   {
	    Queue.wptr = Queue.wptr->next;
	    Queue.wptr->filled = 0;
	    Queue.wptr->gotten = 0;

	    PRINTK("m105: interrupt(): advanced to next write bucket @ 0x%lx\n",(long)Queue.wptr->data);
	   }  

	scanner_on();

	/* signal to client */   
	if (ds.process_sleeps) {
	   if (waitqueue_active(&wq))
              {
	       PRINTK("m105: interrupt(): got bytes, wstack is %ld, will wake up process\n",Queue.wptr->filled);
	       wake_up_interruptible(&wq);
	      } 
	   ds.process_sleeps = 0;
	}

	restore_flags(flags);
}


static int m105_ioctl(struct inode *inode, struct file *file, \
	unsigned int cmd, unsigned long arg)
{
  int r,ss, t;
  long minfo,lt;
  
  PRINTK("m105: ioctl(%d)\n",cmd);
  scanner_standby(); 
  
  if(!arg)
    return -EINVAL;
    
  switch(cmd) {
    case HSIOCGMOD:			/* get modeinfo struct */
      t = _modeindex_;
      ss = SCD_NUM;

      if ((_MODEINFO_.options & SCC_SOFTDETECT) != 0)
         {
          if ((r = m105_get_mode()) < 0)
             {
              _modeindex_ = t;
              SCD_NUM = ss;
/*              return r; */
             } 
         }   

      memcpy_tofs((char*) arg, (char*) &_MODEINFO_, (long) sizeof(struct modeinfo_struct));
      return 0;

    case HSIOCSMOD:			/* set modeinfo struct */
      /*
      memcpy_fromfs((char*) &minfo, (char*) arg, (long) sizeof(long));
       */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
      minfo = get_fs_long(arg);
#else
      get_user(*(long *)&minfo, (long *)arg);
#endif      

       
      PRINTK("m105: set mode to %ld\n",minfo);
      
      if ((minfo > NUM_SCANMODES) || (minfo < 1))
         return -EINVAL;
      
     if (SScanner[SCD_NUM].mode[minfo-1].bpl > 0)
        {
         _modeindex_ = minfo-1;
         _BPL_ = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth); 
         return 0;
        }        
             
      /* supplied mode doesn't match a valid mode */
      return -ENXIO;
/*
      s = sizeof(m105_modeinfo) / sizeof(struct m105_modeinfo_struct);
      printk("about to set custom modeindex to %d/%d\n",modeindex,s); 
#ifdef CUSTOM_MODES
      modeindex = s-2;
      memcpy(&MODEINFO,&minfo,sizeof(struct m105_modeinfo_struct));
      return 0;
#else      
#endif      
*/
    case HSIOCGSCT:
      r = m105_get_mode();
      PRINTK("m105: detected scanner: [%d][%s][%s]\n",SCD_NUM,
                    SScanner[SCD_NUM].type.manufacturer,
                    SScanner[SCD_NUM].type.model);   
                    
      memcpy_tofs((char*) arg, (char*) &(SScanner[SCD_NUM].type), (long) sizeof(struct scanner_type));
      return SCD_NUM;

    case HSIOCGREG:			/* get setting of misc registers */
      memcpy_tofs((char*) arg, (char*) &gi1904_regs, (long) sizeof(struct gi1904_regs_struct));
      return 0;

    case HSIOCSREG:			/* set misc registers */
      memcpy_fromfs((char*) &gi1904_regs, (char*) arg, (long) sizeof(struct gi1904_regs_struct));
      outb_p(gi1904_regs.mr0, GI1904_MISC0);
      outb_p(gi1904_regs.mr1, GI1904_MISC1);
      outb_p(gi1904_regs.mr2, GI1904_MISC2);
      return 0;

    case HSIOCSBUF:			/* set size of internal buffer */
/*      return -ENOSYS;*/
/*    
      if(!(r=get_fs_long(arg)))
	m105_free_mem();
      else {
	if((ss=m105_get_mode()) < 0)
	  return ss;
	else {
	  req_lines = r;
	  m105_get_mem(req_lines);
	}
      }
      return 0;
*/

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
      lt = get_fs_long(arg);
#else
      get_user(*(long *)&lt, (long *)arg);
#endif      
      printk("m105: set bufsize %ld\n",lt);
      return 0;	

    case HSIOCGBUF:			/* get current buffer size */
      printk("m105: get bufsize\n"); 
/*      return -ENOSYS; */
      
      t = 52;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
      put_fs_long((unsigned long) t, arg);
#else      
      put_user(*(int *)&t,(int *)arg);
#endif      
      return 0;

    case HSIOCSSTH:			/* set select(2) threshold */
/*      select_threshold=get_fs_long(arg);*/
      return 0;

    case HSIOCGSTH:			/* get current select(2) threshold */
/*      put_fs_long((unsigned long) select_threshold, arg); */
      return 0;
 
    default:
      return -ENOSYS;
  }
}

static int m105_open(struct inode *inode, struct file *file)
{
	int r;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
	int irq_mask,dma_mask,i;
#endif
        
	PRINTK("m105: open\n");

#ifndef DEBUG_M105
/*
	if (MINOR(inode->i_rdev) != 0)
	   return -ENODEV;
*/
#else
	debug = MINOR(inode->i_rdev);
#endif
	if (ds.device_open)
	   return -EBUSY;

	ds.irq_enabled = 0; 
	ds.process_sleeps = 0;
	outb_p(STANDBY, GI1904_CONTROL);	 /* interrupts and DMA transfers are disabled */

	if ((r = m105_get_mode()) < 0) 
	   {		 /* in standby mode */
	    outb(OFF, GI1904_CONTROL);
	    return r;
	   }
	__cold_sleep(5);

	
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)) 
	if (irq == 0)	/* check for free irq */ 
	   {
	    for (i=0; i<3; i++)
	      {
               if (!(r = request_irq(irq_list[i], m105_interrupt, 
	       			SA_NODEFER|SA_RESTART, M105_ID, M105_ID))) 
	          {
	           PRINTK("m105: using free irq line %d\n",irq_list[i]);
	           irq = irq_list[i];
		   break;
	          }
	      }
	      
	    if (irq == 0)
	       {
	        outb_p(0, GI1904_DMA_IRQ);
		outb(OFF, GI1904_CONTROL); 
		PRINTK("m105: open(): no free irq line\n");
	        return -EBUSY;
	       }	
	   }
	 else
	   {
            if ((r = request_irq(irq, m105_interrupt, SA_INTERRUPT, "M105-Scanner", M105_ID))) 
	       {
	        outb_p(0, GI1904_DMA_IRQ);
		outb(OFF, GI1904_CONTROL); 
		PRINTK("m105: open(): irq %d not free\n",irq);
	        return r;
	       }
	   }    
	     
	switch (irq)
	  {
	   case  3: irq_mask = 0x10; break;
	   case  5: irq_mask = 0x20; break;
	   case 10: irq_mask = 0x08; break;
	   default:
	     outb_p(0, GI1904_DMA_IRQ);
	     outb(OFF, GI1904_CONTROL);
	     free_irq(GI1904_IRQ,M105_ID);
	     printk("m105: open(): wrong irq argument (%d)\n",irq);
	     return -ENODEV;
	  }

	if (dma == 0)	/* check for free dma */ 
	   {
	    for (i=0; i<2; i++)
	      {
	       if (!(r = request_dma(dma_list[i],M105_ID))) 
	          {
	           PRINTK("m105: using free dma channel %d\n",dma_list[i]);
	           dma = dma_list[i];
		   break;
	          }
	      }
	      
	    if (dma == 0)
	       {
	        outb_p(0, GI1904_DMA_IRQ);
		outb(OFF, GI1904_CONTROL);
		free_irq(GI1904_IRQ,M105_ID); 
		PRINTK("m105: open(): no free dma channel\n");
	        return -EBUSY;
	       }	
	   }
	 else
	   {
	    if ((r = request_dma(dma,M105_ID))) 
	       {
	        outb_p(0, GI1904_DMA_IRQ);
		outb(OFF, GI1904_CONTROL);
		free_irq(GI1904_IRQ,M105_ID); 
		PRINTK("m105: open(): dma %d already used\n",dma);
	        return r;
	       }	
	   }  

	switch (dma)
	  {
	   case 1: dma_mask = 0x01; break;
	   case 3: dma_mask = 0x04; break;
	   default:
	     outb_p(0, GI1904_DMA_IRQ);
	     outb(OFF, GI1904_CONTROL);
	     free_irq(GI1904_IRQ,M105_ID);
	     printk("m105: open(): wrong dma argument (%d)\n",dma);
	     return -ENODEV;
	  }
	
	outb_p(irq_mask | dma_mask, GI1904_DMA_IRQ);
	
#else
	outb_p(GI1904_IRQ_MASK | GI1904_DMA_MASK, GI1904_DMA_IRQ); 
	/* set irq & DMA mask _before_ */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,0,33))  
	if ((r = request_irq(GI1904_IRQ, m105_interrupt))) {	     
#else  
	if ((r = request_irq(GI1904_IRQ, m105_interrupt,SA_INTERRUPT,M105_ID,M105_ID))) {	     
#endif  
		/* requesting (and enabling) */
		outb_p(0, GI1904_DMA_IRQ); /* clear irq & DMA mask _before_ */
    		outb(OFF, GI1904_CONTROL); 				     /* the channels */
    		PRINTK("m105: open(): request_irq failed\n");
    		return r;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,0,33))
  if((r = request_dma(GI1904_DMA))) {
#else    
  if((r = request_dma(GI1904_DMA,M105_ID))) {
#endif  

    outb(OFF, GI1904_CONTROL);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,0,33))
    free_irq(GI1904_IRQ);
#else
    free_irq(GI1904_IRQ,M105_ID);
#endif        
    outb_p(0, GI1904_DMA_IRQ); /* clear irq & DMA mask _before_ */
    PRINTK("m105: open(): request_dma %d failed\n",GI1904_DMA);
    return r;
  }

#endif

  scanner_standby(); 
  r = m105_get_mode();
  PRINTK("m105: open mode %d\n",r);
  
  MOD_INC_USE_COUNT;
  ds.device_open = 1;
  return 0;	/* device open, scanner is in standby mode now */
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int m105_release(struct inode *inode, struct file *file)
#else
static void m105_release(struct inode *inode, struct file *file)
#endif
{
  PRINTK("m105: release\n");
/*
  if (buflines)
    m105_free_mem();
 */
  scanner_standby(); 

  outb_p(0, GI1904_DMA_IRQ); /* clear irq & DMA mask _before_ */
  outb(OFF, GI1904_CONTROL);
 
  free_dma(GI1904_DMA);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,0,33))
  free_irq(GI1904_IRQ);
#else
  free_irq(GI1904_IRQ,M105_ID);
#endif    

  ds.device_open = 0;
  MOD_DEC_USE_COUNT;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
  if (auto_irq > 0)
     {
      dma = 0;
      irq = 0; 
     }
           
  return 0;
#endif  
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
static int m105_lseek(struct inode *inode, struct file *file, off_t offset, int origin)
#else
static loff_t m105_lseek(struct file *file, loff_t offset, int origin)
#endif
{
  return -ESPIPE;	/* in case some luser tries to seek on a scanner... */
}

static struct file_operations m105_fops = {
  m105_lseek,
  m105_read,
  NULL,		/* write */
  NULL,		/* readdir */
  m105_select,
  m105_ioctl,
  NULL,		/* mmap */
  m105_open,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0))
  NULL,		/* flush */
#endif
  m105_release,
};

/*  new 2.2.x Interface

        loff_t (*llseek) (struct file *, loff_t, int);
        ssize_t (*read) (struct file *, char *, size_t, loff_t *);
        ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
        int (*readdir) (struct file *, void *, filldir_t);
        unsigned int (*poll) (struct file *, struct poll_table_struct *);
        int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long        
        int (*mmap) (struct file *, struct vm_area_struct *);
        int (*open) (struct inode *, struct file *);
        int (*flush) (struct file *);
        int (*release) (struct inode *, struct file *);
        int (*fsync) (struct file *, struct dentry *);
        int (*fasync) (int, struct file *, int);
        int (*check_media_change) (kdev_t dev);
        int (*revalidate) (kdev_t dev);
        int (*lock) (struct file *, int, struct file_lock *);
*/                                                                

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
static struct miscdevice m105_dev=
{
        240,
        "m105",
        &m105_fops
};

static int get_m105_proc_status(char *buf)
{
   	char *ptr;
        
        ptr = buf;  

   	m105_get_mode();

   	ptr += sprintf(ptr,"scanner type: %s, %s\n",SScanner[SCD_NUM].type.manufacturer,
                                      SScanner[SCD_NUM].type.model);

   	ptr += sprintf(ptr,"auto_irq:     %d\n",auto_irq);
   	ptr += sprintf(ptr,"buffer:       %ld kByte\n",real_bufsize /1024);

   	ptr += sprintf(ptr,"current mode: ");        

#ifdef M105_DEBUG
   	ptr += sprintf(ptr,"0x%02x,  ",_MODEINFO_.detect);        
#endif   	

   	ptr += sprintf(ptr,"%dx%d dpi,",_MODEINFO_.xdpi,_MODEINFO_.ydpi);        
   	ptr += sprintf(ptr,"%dbit,",_MODEINFO_.sc_mode);

        if (_MODEINFO_.depth == 1)
   	   {
	    if (_MODEINFO_.sc_mode == 8) 
	       ptr += sprintf(ptr,"gray\n");        
	     else  
	       ptr += sprintf(ptr,"mono\n");        
	   } 
         else  
   	   ptr += sprintf(ptr,"color\n");        
        
#ifdef M105_DEBUG
   	ptr += sprintf(ptr,"debug:        %d\n",debug);        
#endif   	

        return (ptr - buf);
}

static struct proc_dir_entry proc_m105 = {
        0,                   /* low_ino    */
        sizeof("m105")-1,    /* namelen    */
        "m105",              /* name       */
        S_IFREG | S_IRUGO,   /* mode       */
        1,                   /* nlink      */
        0,                   /* uid        */
        0,                   /* gid        */
        0,                   /* size       */
        NULL,                /* ops        */
        NULL,                /* get_info   */
        NULL,                /* fill_inode */
        NULL,                /* next       */
        NULL,                /* parent     */
        NULL,                /* subdir     */
        NULL,                /* data       */
        get_m105_proc_status,    /* read_proc  */
        NULL                 /* write_proc */
};

#endif                        


int init_module(void)
{
  int n, nodev = 0;

  printk("m105: M105 handheld scanner driver  version 0.1.8\n");

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
   	if (proc_register(&proc_root, &proc_m105) < 0)
           printk("m105: cannot register /proc/m105\n");

   	if (misc_register(&m105_dev) < 0) {
#else
        if (register_chrdev(M105_SCANNER_MAJOR, M105_SCANNER_NAME, &m105_fops)) {
#endif     
            printk("m105: cannot register major number\n");
            return -EIO;
        }

  request_region(GI1904_IOBASE,5,M105_ID);
  
  outb_p(OFF, GI1904_CONTROL);

  for (n=0; n<10000; n++)
    nodev |= (inb_p(GI1904_IDENT) != GI1904_MAGIC);

  if (nodev) {
    printk("m105: GI1904 scanner interface missing or broken\n");
    return -EIO;
  }
  
/*
  if (m105_getmem(DEFAULT_BUFLINES) != 0)
     return -ENOMEM;
 */

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0))    
   	init_waitqueue_head(&wq);
#endif    

   	real_bufsize = init_queue();
        printk("m105: got %ld bytes total buffer space\n",real_bufsize);

        if (real_bufsize < 8192)
           return -ENOMEM;
      
        return 0;
}

void cleanup_module(void)
{
   	done_queue();

   	release_region(GI1904_IOBASE,5);

   	printk("m105: Removing handheld scanner driver from memory\n");

   	if (MOD_IN_USE)
           printk("m105: device is busy, delaying driver removal\n");

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0))
        if (proc_unregister(&proc_root,proc_m105.low_ino) < 0)
           printk("m105: cannot unregister /proc/m105\n");

   	if (misc_deregister(&m105_dev) < 0) 
#else        
        if (unregister_chrdev(M105_SCANNER_MAJOR, M105_SCANNER_NAME))
#endif
           printk("m105: unregister_chrdev() failed\n");
}










/*
static int m105_read(struct inode *inode, struct file *file, char *buf, int
requested)
{
  int r, tmp;

  int BPL;   
  BPL = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth); 

  tmp = requested;

  buflines = get_buflines(); 

  requested -= get_available_bytes(&buf, requested); 

  if (!ds.irq_enabled && (WR_BUFSPC >= BPL)) {
    PRINTK("m105_read: I see some space in my buffer, will enable irq\n");
    scanner_operate();
  }

  if(file->f_flags & O_NONBLOCK) {
    PRINTK("m105_read: reading in O_NONBLOCK mode\n");
    tmp -= requested;
    return tmp;
  }

  while (requested > 0) {
    bytes_missing = requested;
    PRINTK("m105_read: %d bytes are still missing, I'll sleep a bit.
    ds.irq_enabled == %d\n", \
    	bytes_missing, ds.irq_enabled);

    if (!ds.irq_enabled) {
      PRINTK("m105_read: re-enabling interrupts\n");
      scanner_operate();
    }

    ds.process_sleeps = 1;
    interruptible_sleep_on(&wq);
    requested -= get_available_bytes(&buf, requested);

    if (SIGNAL_OCCURED)
       {
        tmp -= requested;
        return tmp;
       }  
  }

  return tmp;
}

*/

/*
#ifdef OLD_KERNEL
static void m105_interrupt(int unused_pointer_to_pt_regs_struct)
#else
static void m105_interrupt(int irq_dummy,void *dev_id,struct pt_regs
*unused_pointer_to_pt_regs_struct)
#endif
{
  static int bytes_til_wakeup;
  int BPL;
  
  BPL = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth);

  if (!ds.irq_enabled) {
    PRINTK("scanner.c: m105_interrupt: unexpected interrupt\n");
    return;
  }

  if (!dma_terminal_count(GI1904_DMA)) {
    PRINTK("scanner.c: m105_interrupt: DMA transfer not finished yet\n");
    return;
  }

  PRINTK("m105_interrupt: I have been called, bytes_missing == %d\n",
  bytes_missing);

  scanner_standby();

  if (bytes_missing > 0) {
    bytes_til_wakeup = bytes_missing;
    bytes_missing = 0;
  }
  
  wr.index += BPL;
  if(wr.index >= WR_BLKEND) {
    wr.blkptr = wr.blkptr->next;
    if(!wr.blkptr) {
      wr.blkptr = memroot;
      wr.index -= bufsiz;
    }
  }
  
  if (wr.index == rd.index)
    ds.buffer_empty = 0;

  PRINTK("m105_interrupt: wr.index is now %d\n", wr.index);

  if (bytes_til_wakeup > 0)
    bytes_til_wakeup -= BPL;

  if (bytes_til_wakeup <= 0)
    if(ds.process_sleeps) {
      PRINTK("m105_interrupt: the requested bytes are there, will wake up
      process\n");
      wake_up_interruptible(&wq);
      ds.process_sleeps = 0;
    }

  if (WR_BUFSPC >= BPL) {
    PRINTK("m105_interrupt: there's still free buffer space (%d bytes), irq
    stays enabled\n", \
    	WR_BUFSPC);

    scanner_operate();
  } else {
    PRINTK("m105_interrupt: buffer is full -- turning off irq\n");
    ds.irq_enabled = 0;
    if(ds.process_sleeps) {
      wake_up_interruptible(&wq);
      ds.process_sleeps = 0;
    }
  }
  return;
}
*/
/*
static __inline__ void __scanner_operate(void)
{
  cli();
  setup_xfer(WR_ADDR);
  ds.irq_enabled = 1;
  sti();

//  outb(0xff,GI1904_MISC0); *
//  enable_dma(GI1904_DMA);
   
  outb(OPERATE, GI1904_CONTROL);
}
*/    

/*
static void __m105_free_mem(void)
{
  struct buffer_block_struct *tmp = memroot, *next;
  unsigned int size,i;

  if (!buflines) {
    PRINTK("m105_free_mem: there's nothing to free\n");
    return;
  }
  ds.irq_enabled = 0;

  scanner_standby();	// * make sure DMA xfers are really off 

  while (tmp) {
    size = (tmp->blksiz);
    
    i=0;
    while ((km_bucket[i] < size) && (i<5))
      i++;
    
    size = km_bucket[i];
    next = tmp->next;

    PRINTK("about to kfree_s'd %d bytes (data+header) at 0x%010lx/0x%010lx\n",
    size, (long) tmp->data,(long)tmp);

    kfree_s((char*) tmp->data, size); 
    kfree_s((char*) tmp, HEADERSIZ ); 
// *    kfree((char *)tmp); *
    
    tmp = next;
  };
  
  memroot = NULL; 
  rd.blkptr = NULL; 
  wr.blkptr = NULL;
  buflines = 0; 
  bufsiz = 0;
}
*/


/*
static int __m105_get_mem(int lines)
{
  struct buffer_block_struct *tmp, **lastptr;
  unsigned int blksiz;
  int BPL,buf_lin,k;


  BPL = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth);

  if ((lines * BPL) > MAX_BUFSIZ)
    return -EINVAL;

  if (buflines)
    __m105_free_mem();

  lastptr = &memroot;

  while (lines > 0) {

//    blksiz = (((lines * MODEINFO.bpl) > MAX_BLK) ? \
//	(MAX_BLK - MAX_BLK % MODEINFO.bpl) : (lines * MODEINFO.bpl));
  
    k = 5;
    
    buf_lin = km_bucket[k] / BPL;

    while ((buf_lin > lines) && (k > 0))
      {
       k -= 1; 
       buf_lin = km_bucket[k] / BPL;
      } 
        
// *    blksiz = (MODEINFO.bpl*MODEINFO.sc_mode*MODEINFO.depth) * 2; * / // *
get sufficient buffer for 2 lines *
    
    blksiz = BPL * buf_lin; 

    PRINTK("used km_bucket[%d]:%d   buflines:%d\n",k,km_bucket[k],buf_lin);
     
    tmp = (struct buffer_block_struct*)kmalloc(HEADERSIZ , GFP_DMA);

    if (!tmp)
       {
        __m105_free_mem();
        return -ENOMEM;
       }

    tmp->data = (char *)kmalloc(blksiz, GFP_DMA);
        // * get memory and a little safety offset for DMA transfers *
        
    if (tmp->data == NULL) {
      __m105_free_mem();
      return -ENOMEM;
    }
    
    tmp->next = NULL;
    *lastptr = tmp;
    lastptr = &(tmp->next);
    tmp->blksiz = blksiz;
    tmp->blkoff = buflines * BPL;

    lines -= blksiz / BPL;
    buflines += blksiz / BPL;
    bufsiz += blksiz;
    
    PRINTK("kmalloc'ed %d bytes (data+header) at 0x%010lx/0x%010lx, %d lines,
    remaining: %d, offset: %d\n", \
    	blksiz+HEADERSIZ, (long) tmp->data, (long)tmp,buflines, lines,
    	tmp->blkoff);
    
    rd.blkptr = memroot; 
    rd.index = 0;
    
    wr.blkptr = memroot; 
    wr.index = 0;
    
    ds.buffer_empty = 1;
  }
 
  return 0;
}
*/



/*
static void m105_freemem(void)
{
  if (bufsiz <= 0) {
    PRINTK("m105_free_mem: there's nothing to free\n");
    return;
  }
  
  ds.irq_enabled = 0;
  scanner_standby();	// * make sure DMA xfers are really off *

  kfree_s((char*) memroot->data, memroot->blksiz); 
  kfree_s((char*) memroot, HEADERSIZ ); 
  
  memroot = NULL; 
  rd.blkptr = NULL; 
  wr.blkptr = NULL;
  buflines = 0; 
  bufsiz = 0;
}
*/

/*
static int get_buflines(int chunk_size,int BPL)
{
   int BPL;
   
   BPL = (_MODEINFO_.bpl*_MODEINFO_.sc_mode*_MODEINFO_.depth);

   if (BPL > 0)
     return (chunk_size / BPL);   
     
   return DEFAULT_BUFLINES;  
}
*/

/*
struct buffer_block_struct *get_memblock(struct buffer_block_struct *prev, int
chunk_size)
{
   struct buffer_block_struct *buff;
   int i;
   
   buff = (struct buffer_block_struct*)kmalloc(HEADERSIZ , GFP_DMA);
   if (buff == NULL)
      return NULL;
      
   buff->data = (char*)kmalloc(chunk_size , GFP_DMA);
   if (buff->data == NULL)
      {
       kfree((char*)buff);
       return NULL;
      }    
      
   buff->blksiz = chunk_size;
   buff->blkoff = 0;
   buff->blklines = get_buflines(chunk_size);   
   buff->next = prev;
   
   return buff;
}

int count_buflines(struct buffer_block_struct *blk)
{
   int ln;
   
   if (blk == NULL)
      return 0;
   
   ln = 0;
      
   if (blk->next != NULL)
      ln = count_buflines(blk->next);
   
   ln += get_buflines(blk->blksiz);
   
   return ln;      
}

static int m105_getmem(int lines)
{
   struct buffer_block_struct *tmp,*prev;
   bufsiz = 65512;

   prev = NULL;
   i = 4;
   while (i > 0)
     {
      i--;
      tmp = get_memblock(prev,km_bucket[i]);
   
      while (tmp != NULL)
        {
         prev = tmp;
         tmp = get_memblock(prev,km_bucket[i]);
        } 
     } 
      
   memroot = prev;
   buflines = count_buflines(memroot);

   rd.blkptr = memroot; 
   rd.index = 0;
    
   wr.blkptr = memroot; 
   wr.index = 0;
    
   ds.buffer_empty = 1;
  
   return 0;
}
*/

/*     
static __inline__ int get_available_bytes(char **user_buf, int requested)
{
  int bytes_to_copy;
  int chunksiz, bytes_copied = 0;
  
  cli(); 
  bytes_to_copy = MIN(requested, RD_BUFSPC); 
  sti();
  
  PRINTK("get_available_bytes: %d bytes to copy (%d requested, RD_BUFSPC ==
  %d)\n", \
  	bytes_to_copy, requested, RD_BUFSPC);
  
  PRINTK("get_available_bytes: rd.index == %d, wr.index == %d\n", rd.index,
  wr.index);
  
  while(bytes_to_copy > 0) {
    cli(); 
    chunksiz = MIN(bytes_to_copy, RD_BLKEND - rd.index); 
    sti();
    
    PRINTK("get_available_bytes: copying chunk of %d bytes to user space\n",
    chunksiz);
    PRINTK("get_available_bytes: RD_BLKEND is %d\n", RD_BLKEND);
    
    memcpy_tofs(*user_buf, RD_ADDR, (long) chunksiz);
    *user_buf += chunksiz;
    rd.index += chunksiz;
    bytes_copied += chunksiz;
    bytes_to_copy -= chunksiz;
    
    if(rd.index >= RD_BLKEND) {
      rd.blkptr = rd.blkptr->next;
      if(!rd.blkptr) {	// * last block in linked list *
	rd.blkptr=memroot;
	rd.index -= bufsiz;
      }
    }
  }
  cli(); 
  
  if (rd.index == wr.index) 
     ds.buffer_empty = 1; 
  
  sti();
  
  PRINTK("get_available_bytes: copied total of %d bytes to user space\n",
  bytes_copied);
  
  return bytes_copied;
}
*/
