






//#include "utils.h"

#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

#include "t_scanner.h"
#include "utils.h"
#include "copyright.h"

#define ErrorMsg(x)  perror(x)

t_scanner_mode scanner_mode[] = { 
                                  { 1, "100/ bw"}
                                 ,{ 2, "200/ bw"}
                                 ,{ 3, "300/ bw"}
                                 ,{ 4, "400/ bw"}

                                 ,{ 5, "100/ gray"}
                                 ,{ 6, "200/ gray"}
                                 ,{ 7, "300/ gray"}
                                 ,{ 8, "400/ gray"}


                                 ,{ 9, "100/ color24"}
                                 ,{10, "200/ color24"}
/*
                                 ,{11, "300/ color24"}
                                 ,{12, "400/ color24"}
*/
                                 ,{11, "100/ color12"}
                                 ,{12, "200/ color12"}
/*                                 ,{11, "300/ color12"}
                                 ,{12, "400/ color12"}
*/
                                 ,{ 0, "unknown"}
                                };

// int FD_Ready(int fd)
// {
//   struct fd_set Read_Set;
//   struct timeval TimeOut;
// 
// //  FD_ZERO(&Read_Set);
//   FD_SET(fd,&Read_Set);
//   TimeOut.tv_sec=0;
//   TimeOut.tv_usec=0;
// 
// //#ifdef hpux
// //  select(fd+1,(int *)&Read_Set,0,0,&TimeOut);
// //#else
//   select(fd+1,&Read_Set,0,0,&TimeOut);
// //#endif    
//   if FD_ISSET(fd,&Read_Set)
//        return 1;
//     else
//        return 0;   
// }


		
void t_scanner_image::Init(int w, int h, int d)
{
	width = w;
	height = h;
	depth = d;

	bpl = (width * depth) / 8;
	bpp = depth / 8;
	size = bpl * height;
	
	buffer = new unsigned char [size];
}

void t_scanner_image::Cut(int x1, int y1, int x2, int y2)
{
	if ((x2 < x1) || (y2 < y1))
	   {
	    fprintf(stderr,"wrong coords for cut image\n");	
	    return;
	   } 

	int w = __min(x2,width) - __max(0,x1);
	int h = __min(y2,height) - __max(0,y1);
	
	w += 4;
	w >>= 2;
	w <<= 2;
	
	t_scanner_image *N = new t_scanner_image;
	N->Init(w,h,depth);
	
	for (int i=0; i<h; i++)
	  for (int j=0; j<w; j++)
	    {
	     t_color_3i sc;
	     if (GetPixel_(j+x1,i+y1,&sc) == false)
		    break;
			
	     if (N->SetPixel_(j,i,&sc) == false)
		    break;
	    }

	unsigned char *swap = buffer;
	buffer = N->buffer;
	N->buffer = swap;

#ifdef DEBUG
	fprintf(stderr,"t_scanner_image::Cut(%d,%d,%d,%d) ow:%d,oh:%d\tw:%d,h:%d\n",x1,y1,x2,y2,width,height,w,h);
#endif	

	width = w;
	height = h;

	bpl = (width * depth) / 8;
	size = bpl * height;

	
	N->Done();	
	delete N;
	N = NULL;
	
}


#ifdef DEBUG
void t_scanner_do_break(void)
{
	fprintf(stderr,"t_scanner_do_break\n");
}
#endif

void t_scanner_image::Resize(int w, int h, int d)
{
#ifdef DEBUG
	fprintf(stderr,"t_scanner_image::Resize(%d,%d,%d)\n",w,h,d);
#endif	
//	t_scanner_do_break();

	if ((width == w) && (depth == d) && (height == h))
	   return;
	   
	Done();
	    	
//	t_scanner_do_break();

	width = w;
	height = h;
	depth = d;

	bpl = (width * depth) / 8;
	bpp = depth / 8;
	size = bpl * height;

//	t_scanner_do_break();

	buffer = new unsigned char [size];
}

void t_scanner_image::Done(void)
{
	delete buffer;
	buffer = NULL;
	
	width = 0;
	height = 0;
	depth = 0;
	size = 0;
}


#ifdef DEBUG
void dummy_break(void)
{
}
#endif

char t_scanner_image::GetPixel_(int x, int y, t_color_3i *color)
{
#ifdef DEBUG
	dummy_break();
#endif

	color->red = 0;	
	color->green = 0;	
	color->blue = 0;	
	
	if ((x > width) || (y > height))
	   return false;	

	int bpos = (x*bpp) + (y*bpl);
	int boff,mask,kv;
	double new_color;
	
	switch (depth)
	  {
	   case 1: // 1bit bitmapped black/white
	     bpos = (x / 8) + (y*bpl);
	     boff = 7 - (x % 8);  
	     mask = (1 << boff);
	     
	     if ((buffer[bpos] & mask) > 0)
	     	{
		 color->red = 255;	
		 color->green = 255;	
		 color->blue = 255;	
		}
	      else	
	     	{
		 color->red = 0;	
		 color->green = 0;	
		 color->blue = 0;	
		}
	     break;
	    
	   case 8: // 8bit-grayscale
	     color->red = buffer[bpos];
	     color->green = buffer[bpos];
	     color->blue = buffer[bpos];
	     break;
	     
	   case 24: // 8bit-truecolor 
	     color->red   = buffer[bpos];
	     color->green = buffer[bpos+1];
	     color->blue  = buffer[bpos+2];
	     break;

	   case 12: // 4bit-truecolor 
	     bpos = ((x >> 1) * 3) + (y*bpl);

		 unsigned char A,B,C,C1,C2,C3,C4,C5,C6;
		 
		 A = buffer[bpos];
		 B = buffer[bpos+1];	
		 C = buffer[bpos+2];	
		 
		 C1 = (A & 0xf0);
		 C2 = (A & 0x0f);
		 C2 <<= 4;
		 C3 = (B & 0xf0);

		 C4 = (B & 0x0f);
		 C4 <<= 4;

		 C5 = (C & 0xf0);
		 C6 = (C & 0x0f);
		 C6 <<= 4;
		  
	     if ((x % 2) == 0)
		    {
			 color->red   = C1;
 	    	 color->green = C2;
 	    	 color->blue  = C3;
			} 
		  else	
		    {
			 color->red   = C4;
	    	 color->green = C5;
	    	 color->blue  = C6;
			} 
		 break;
		 
	   case 18: // 6bit-truecolor 
	   default:
	     fprintf(stderr,"t_Scanner_Image::GetPixel() with depth %d not supported\n",depth);
		 return false;
	     break;
	  }
	
	return true;		
}

char t_scanner_image::SetPixel_(int x, int y, t_color_3i *color)	
{
	int bpos = (x*bpp) + (y*bpl);
	int boff,mask;
	double new_color;
	
	switch (depth)
	  {
	   case 1: // 1bit bitmapped black/white
	     bpos = (x / 8) + (y*bpl);
	     boff = 7 - (x % 8);  
	     mask = (1 << boff);
	     
	     if ((color->red + color->green + color->blue) > 0)
	     	buffer[bpos] |= mask;
	      else
	        buffer[bpos] &= ~mask;		
	     break;
	    
	   case 8: // 8bit-grayscale
	     new_color =   (double)color->green * 0.59;
	     new_color += ((double)color->red * 0.30);
	     new_color += ((double)color->blue * 0.11);
//	     new_color /= 256.0;
	     
	     buffer[bpos] = (unsigned char)new_color;
	     break;
	     
	   case 24: // 8bit-truecolor 
	     buffer[bpos]   = color->red;
	     buffer[bpos+1] = color->green;
	     buffer[bpos+2] = color->blue;
	     break;        		

	   case 12: // 4bit-truecolor 
	     bpos = ((x >> 1) * 3) + (y*bpl);

//		 unsigned char A,B,C,C1,C2,C3,C4,C5,C6;
// 		 A = buffer[bpos];
// 		 B = buffer[bpos+1];	
// 		 C = buffer[bpos+2];	
// 		 
// 		 C1 = (A & 0xf0);
// 		 C2 = (A & 0x0f);
// 		 C2 <<= 4;
// 		 C3 = (B & 0xf0);
// 
// 		 C4 = (B & 0x0f);
// 		 C4 <<= 4;
// 
// 		 C5 = (C & 0xf0);
// 		 C6 = (C & 0x0f);
// 		 C6 <<= 4;
		  
	     if ((x % 2) == 0)
		    {
			 buffer[bpos  ] = (color->red   & 0xf0) | (buffer[bpos  ] & 0x0f); // C1;
 	    	 buffer[bpos  ] = (color->green   >> 4) | (buffer[bpos  ] & 0xf0); // C2;
			 buffer[bpos+1] = (color->blue  & 0xf0) | (buffer[bpos+1] & 0x0f); // C3;
			} 
		  else	
		    {
 	    	 buffer[bpos+1] = (color->red     >> 4) | (buffer[bpos+1] & 0xf0); // C4;
			 buffer[bpos+2] = (color->green & 0xf0) | (buffer[bpos+2] & 0x0f); // C5;
 	    	 buffer[bpos+2] = (color->blue    >> 4) | (buffer[bpos+2] & 0xf0); // C6;
			} 
	     break;
		 
	   case 18: // 6bit-truecolor 
	   default:
	     fprintf(stderr,"t_Scanner_Image::SetPixel() with depth %d not supported\n",depth);
		 return false;	
	     break;
	  }	
	  
	return true;	  	
}



char t_scanner::Init(char *dev)
{
   working = false;
   bptr = 0;
   __Image = NULL;
//   img_size = 0;
   max_lines = 480;
   
   strcpy(devnm,dev);
   
   fd = open(devnm,O_RDONLY|O_NONBLOCK); 
   
   if (fd < 1)
      return false;
  
   bool b = GetMode();
   int d = GetType();

   return true;
}


int t_scanner::GetType(void)
{
   if (fd < 1)
      fd = open(devnm,O_RDONLY|O_NONBLOCK|O_NDELAY);
      
   if (fd < 1)
      return false;   
       
   int i = ioctl(fd,HSIOCGSCT,&type);

#ifdef DEBUG
   fprintf(stderr,"scanner type result: %d\n",i);
#endif
   
   return i;
}

char t_scanner::GetMode(void)
{
   
   if (fd < 1)
      fd = open(devnm,O_RDONLY|O_NONBLOCK);
      
   if (fd < 1)
      return false;   
       
//   GetType(); 
      
   if (ioctl(fd,HSIOCGMOD,&mode) == -1)
      {
       ErrorMsg("get scanner mode");
       return false;
      } 

//#ifdef DEBUG
   fprintf(stderr,"scanner: dpi:%d\tmode:%d\tdepth:%d\tbits/line:%d\n",
             mode.xdpi, mode.sc_mode, mode.depth, mode.bpl*8);    
//#endif

   _Depth = mode.depth * mode.sc_mode;
   _Width = (mode.bpl * 8); 

   if (__Image != NULL)
      __Image->Resize(_Width,max_lines,_Depth);
//    else
//      fprintf(stderr,"WARNING! NULL pointer\n");  

//   close(fd);
//   fd = -1;
  
   return true;      
}

void t_scanner::Resize(void)
{
   fprintf(stderr,"t_scanner::Resize() not implemented\n");
}


void t_scanner::Start(int mod)
{
   // sc_mode == bits per pixel
   // depth   == color planes
   // bpl     == (bits per line / 8)
   int old_m = mode_num;
   
   if (fd < 1)
      {
       fd = open(devnm,O_RDONLY|O_NONBLOCK);
       SetMode(mod);
      } 

   if (fd < 1)
      return;
       
//   SetMode(mode_num);
/* 
   BPL = mode.sc_mode * mode.depth * mode.bpl;
   width = BPL * 8 / (mode.depth * mode.sc_mode);
 */
   if (GetMode() == false)
      return;
   
   if (mode_num != old_m) 
      Resize();
  
   bptr = 0;
   scanned_size = 0;

   if (__Image != NULL)
      {
       __Image->Done();	
       delete __Image;
       __Image = NULL;
      } 

	__Image = new t_scanner_image;
	__Image->Init(_Width,max_lines,_Depth); // buffer = new unsigned char [img_size];

//   	fprintf(stderr,"t_scanner::Start(%d): w:%d\n",mod,__Image->width);
//   	if (__Image != NULL)
	working = true;
}


bool t_scanner::Callback(void)
{
	if (working == false)
	   return false;

	if (fd < 1)
	   return false;
         
//   if ((_Image == NULL) || (img_size == 0))
	if (__Image == NULL)
	   {	   	
	    working = false;
       	    return false;
	   } 

	int RD = 0;

//   int rc = 0;           
	while (FD_Ready(fd) == true)
	  {
//       if (rc > 0)
//         fprintf(stderr,"rc %d\n",rc);
          
//       rc++;
       long int gt = __Image->size - bptr;
       
       if (gt < __Image->bpl)
          {
           fprintf(stderr,"no idea gt:%ld, img_size:%ld, bptr:%ld\n",gt,__Image->size,bptr);
           exit(1);
           return false;
          } 

       int rd = read(fd,&__Image->buffer[bptr],gt);

       if (rd > 0)
          {
           bptr += rd;
           RD += rd;
       
           scanned_size = bptr;
           _Height = bptr / __Image->bpl;
           
//           if (bptr == __Image->size) 
	   if (_Height >= __Image->height)
              {
	       fprintf(stderr,"not enough memory allocated, try higher matrix values\n");
	       fprintf(stderr,"bptr:%ld,img_size:%ld\th:%d,img_height:%d\n",bptr,__Image->size,_Height,__Image->height);		       
               Stop();
               return true;   
              } 
          } 
      }      
   
   if (RD > 0)
      {
#ifdef DEBUG
       fprintf(stderr,"scanner read %d bytes\n",RD);
#endif       
       return true;
      } 
   
   return false;
}

void t_scanner::Stop(void)
{
   char st[256];
     
   
   if (fd < 1)
      return;
   
   // discard 
   while (FD_Ready(fd) == true)
      read(fd,&st[0],240);
              
   working = false;   
//   Image = NULL;
   bptr = 0;
   __Image->height = _Height;
   
   close(fd);
   fd = -1;
}
 
int t_scanner::SetMode(int m)
{
   long mmm;
 
   mmm = m;
   
   if (ioctl(fd,HSIOCSMOD,&mmm) == 0)
      {
       /*
       BPL = (mode.bpl * mode.sc_mode * mode.depth);
       width = (BPL * 8) / (mode.depth * mode.sc_mode);
        */
       if (GetMode() == false)
          return -m; 
        
       mode_num = m;
       return m;
      } 
    else
      {
       ErrorMsg("scanner: HSIOCSMOD");  
//       memcpy(&mode,&modeinfo[i],sizeof(modeinfo_struct));
      } 
     
   return -m;  
}




void t_scanner::Done()
{
   if (fd > 1)
      close(fd);
      
   fd = -1;   

//   if ((img_size != 0) && (_Image != NULL))
   if (__Image != NULL)
      {
       __Image->Done();
       delete __Image;
       __Image = NULL;
       scanned_size = 0;
       bptr = 0;
      } 
}




void t_scanner::Rotate(void)
{
#ifdef DEBUG
//   fprintf(stderr,"rotate: w: %d, h:%d\n",Width,Height);
#endif   
   if (__Image == NULL)
      return;
      	 

   fprintf(stderr,"w:%d,h:%d\tW:%d,H:%d\n",__Image->width,__Image->height,_Width,_Height);	

   switch (_Depth)
      {
       case 1: 
       case 8:
	   case 12:
       case 24:
       	 {
	  int w = __Image->width;
	  int h = __Image->height;
	  
          while ((h % 4) != 0  )
             h++;

 	  t_scanner_image *ni = new t_scanner_image;
 	  ni->Init(h,w,_Depth);
	
	  for (int y=0; y<h; y++)
	    for (int x=0; x<w; x++)
	       {
	        t_color_3i SC;
	     
			if (__Image->GetPixel_(x,y,&SC) == false)
				break;
				
	        if (ni->SetPixel_(y,(w - x) -1,&SC) == false)
				break;
	       }
	  
	  __Image->Done();
	  delete __Image;
          __Image = ni;
          
//          _Height = _Width;
//          _Width = ni->width;
	 }
         break;	

       default:
         fprintf(stderr,"rotate image with depth %d probably not supported\n",_Depth);
         break;    
      }
}


void t_scanner::Mirror(void)
{
#ifdef DEBUG
//   fprintf(stderr,"rotate: w: %d, h:%d\n",Width,Height);
#endif 
	if (__Image == NULL)
	   return;
	   	 	  
   switch (_Depth)
      {
       case 1: 
       case 8:
	   case 12:
       case 24:
       	 {
	  int w = __Image->width;
	  int h = __Image->height;
	  
 	  t_scanner_image *ni = new t_scanner_image;
 	  ni->Init(w,h,_Depth);
	
	  for (int y=0; y<h; y++)
	    for (int x=0; x<w; x++)
	       {
	        t_color_3i SC;
	     
			if (__Image->GetPixel_(x,y,&SC) == false)
				break;
				
	        if (ni->SetPixel_((w - x) -1,y,&SC) == false)
				break;
	       }
	  
	  __Image->Done();
	  delete __Image;
          __Image = ni;
	 }
         break;	

       default:
         fprintf(stderr,"mirror image with depth %d probably not supported, check t_scanner_image::[GS]etPixel() for details\n",_Depth);
         break;    
      }
}
